import * as v from 'valibot';
import {BaseDataSource, GenerationStrategy} from './data-source';

export interface ResourceIdConfig<IdSchema extends v.BaseSchema, DataSource extends BaseDataSource = BaseDataSource> {
	generationStrategy: GenerationStrategy<DataSource>;
	serialize: (id: unknown) => string;
	deserialize: (id: string) => v.Output<IdSchema>;
	schema: IdSchema;
}

export interface ResourceState<
	ItemName extends string = string,
	RouteName extends string = string,
	IdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
> {
	idAttr: IdAttr;
	itemName: ItemName;
	routeName: RouteName;
	idConfig: ResourceIdConfig<IdSchema>;
	fullTextAttrs: Set<string>;
	canCreate: boolean;
	canFetchCollection: boolean;
	canFetchItem: boolean;
	canPatch: boolean;
	canEmplace: boolean;
	canDelete: boolean;
}

export interface Resource<
	Schema extends v.BaseSchema = v.BaseSchema,
	CurrentName extends string = string,
	CurrentRouteName extends string = string,
	CurrentIdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
> {
	dataSource?: unknown;
	schema: Schema;
	state: ResourceState<CurrentName, CurrentRouteName, CurrentIdAttr, IdSchema>;
	id<NewIdAttr extends CurrentIdAttr, TheIdSchema extends v.BaseSchema>(
		newIdAttr: NewIdAttr,
		params: ResourceIdConfig<TheIdSchema>
	): Resource<Schema, CurrentName, CurrentRouteName, NewIdAttr, TheIdSchema>;
	fullText(fullTextAttr: string): this;
	name<NewName extends CurrentName>(n: NewName): Resource<Schema, NewName, CurrentRouteName, CurrentIdAttr, IdSchema>;
	route<NewRouteName extends CurrentRouteName>(n: NewRouteName): Resource<Schema, CurrentName, NewRouteName, CurrentIdAttr, IdSchema>;
	canFetchCollection(b?: boolean): this;
	canFetchItem(b?: boolean): this;
	canCreate(b?: boolean): this;
	canPatch(b?: boolean): this;
	canEmplace(b?: boolean): this;
	canDelete(b?: boolean): this;
}

export const resource = <
	Schema extends v.BaseSchema,
	CurrentName extends string = string,
	CurrentRouteName extends string = string,
	CurrentIdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
>(schema: Schema): Resource<Schema, CurrentName, CurrentRouteName, CurrentIdAttr, IdSchema> => {
	const resourceState = {
		fullTextAttrs: new Set<string>(),
		canCreate: false,
		canFetchCollection: false,
		canFetchItem: false,
		canPatch: false,
		canEmplace: false,
		canDelete: false,
	} as ResourceState<CurrentName, CurrentRouteName, CurrentIdAttr, IdSchema>;

	return {
		get state(): ResourceState<CurrentName, CurrentRouteName, CurrentIdAttr, IdSchema> {
			return Object.freeze({
				...resourceState
			}) as unknown as ResourceState<CurrentName, CurrentRouteName, CurrentIdAttr, IdSchema>;
		},
		canFetchCollection(b = true) {
			resourceState.canFetchCollection = b;
			return this;
		},
		canFetchItem(b = true) {
			resourceState.canFetchItem = b;
			return this;
		},
		canCreate(b = true) {
			resourceState.canCreate = b;
			return this;
		},
		canPatch(b = true) {
			resourceState.canPatch = b;
			return this;
		},
		canEmplace(b = true) {
			resourceState.canEmplace = b;
			return this;
		},
		canDelete(b = true) {
			resourceState.canDelete = b;
			return this;
		},
		id<NewIdAttr extends CurrentIdAttr, NewIdSchema extends IdSchema>(newIdAttr: NewIdAttr, params: ResourceIdConfig<NewIdSchema>) {
			resourceState.idAttr = newIdAttr;
			resourceState.idConfig = params;
			return this as Resource<Schema, CurrentName, CurrentRouteName, NewIdAttr, NewIdSchema>;
		},
		newId(dataSource: BaseDataSource) {
			return resourceState?.idConfig?.generationStrategy?.(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (
				schema.type === 'object'
				&& (
					schema as unknown as v.ObjectSchema<
						Record<string, v.BaseSchema>,
						undefined,
						Record<string, string>
					>
				)
					.entries[fullTextAttr]?.type === 'string'
			) {
				resourceState.fullTextAttrs?.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name<NewName extends CurrentName>(n: NewName) {
			resourceState.itemName = n;
			return this as Resource<Schema, NewName, CurrentRouteName, CurrentIdAttr, IdSchema>;
		},
		route<NewRouteName extends CurrentRouteName>(n: NewRouteName) {
			resourceState.routeName = n;
			return this as Resource<Schema, CurrentName, NewRouteName, CurrentIdAttr, IdSchema>;
		},
		get idAttr() {
			return resourceState.idAttr;
		},
		get itemName() {
			return resourceState.itemName;
		},
		get routeName() {
			return resourceState.routeName;
		},
		get schema() {
			return schema;
		},
	} as Resource<Schema, CurrentName, CurrentRouteName, CurrentIdAttr, IdSchema>;
};
